import torch as th


class Box:
    def __init__(self, dim, low=None, high=None, dtype=th.float):
        """Constructor for Space.

        :param dim: Number of dimension in of the Box.
        :type dim: int
        :param low: Lower bounds of the Box space.
        :type low: float or list or th.Tensor
        :param high: Higher bounds of the Box space.
        :type high: float or list or th.Tensor
        :param dtype: Data type used in the Space, must be explicitly provided.
        :type dtype: th.dtype
        """
        assert dtype is not None, "Data type must be explicitly provided."
        assert isinstance(
            dtype, th.dtype
        ), "Data type must be of class `torch.dtype`."
        self.dtype = dtype

        assert dim > 0, "Dimension must be a strictly positive integer."
        self.__dim = dim
        self.__shape = th.Size([self.dim])

        if low is None:
            self.low = th.full(self.shape, -float("Inf"))
        elif th.as_tensor(low).ndim == 0:  # if scalar
            self.low = th.full(self.shape, low)
        else:
            assert (
                low.shape == self.shape
            ), "Lower boundary must have same dimensions as space Box."
            self.low = th.as_tensor(low)
        if high is None:
            self.high = th.full(self.shape, float("Inf"))
        elif th.as_tensor(high).ndim == 0:
            self.high = th.full(self.shape, high)
        else:
            assert (
                high.shape == self.shape
            ), "Higher boundary must have same dimensions as space Box."
            self.high = th.as_tensor(high)

    @property
    def dim(self):
        """Returns the dimension of the Box.

        .. note::
            This is **not** equivalent to the tensor dimension.

        :return: The number of dimensions in the Box.
        :rtype: int
        """
        return self.__dim

    @property
    def shape(self):
        """Returns the shape of the Box.

        :return: The PyTorch dimensions of the Box.
        :rtype: th.Size
        """
        return self.__shape



import torch as th


class BaseModel:
    """Base class for forward models used in the DISCO framework.

    .. note::
        Unlike Gym, this is *not* an environment. Effectively, this means
        that the model does not hold the current system state. This is useful to
        compute several system trajectories in parallel and vectorize the code.

        Furthermore, as we are interested in uncertain dynamics, models should
        hold a probability distribution over their uncertain parameters. This
        means each derived model must support the BaseModel functions to assign
        and sample from a distribution over some or all of its parameters.
    """

    def __init__(
        self, dt=0.05, uncertain_params=None, params_dist=None,
    ):
        """Constructor for BaseModel.

        :param dt: Duration of each discrete update in s. (default: 0.05)
        :type dt: float
        :param uncertain_params: A list containing the uncertain parameters of
            the forward model. Is used as keys for assigning sampled parameters
            from the `params_dist` function.
        :type uncertain_params:
        :param params_dist: A distribution to sample parameters for the forward
            model.
        :type params_dist: torch.distributions.distribution.Distribution
        """
        assert dt > 0, "Delta t must be greater than zero."
        self.__dt = dt

        assert (uncertain_params is None and params_dist is None) or (
            uncertain_params is not None and params_dist is not None
        ), "Need to specify uncertain parameters and their distribution."
        self.__params_keys = uncertain_params
        self.params_dist = params_dist

    @property
    def dt(self):
        """Returns the discrete timestep duration."""
        return self.__dt

    @property
    def uncertain_params(self):
        """Returns the list of uncertain parameters."""
        return self.__params_keys

    @property
    def params_dist(self):
        """Returns the distribution over uncertain parameters."""
        return self.__params_dist

    @params_dist.setter
    def params_dist(self, dist):
        if dist is not None:
            assert len(dist.mean) == len(
                self.__params_keys
            ), "Number of uncertain parameters must match their distribution."
        self.__params_dist = dist

    @property
    def action_space(self):
        """Returns a Space object."""
        raise NotImplementedError

    @property
    def observation_space(self):
        """Returns a Space object."""
        raise NotImplementedError

    def step(self, states, actions, sampled_params=None):
        """Receives tensors of current states and actions and computes the
        states for the subsequent timestep. If sampled parameters are provided,
        these must be used, otherwise revert to mode of distribution over the
        uncertain parameters.

        Must be bounded by observation and action spaces.

        :param states: A tensor containing the current states of one or multiple
            trajectories.
        :type states: th.Tensor
        :param actions: A tensor containing the next planned actions of one or
            multiple trajectories.
        :type actions: th.Tensor
        :param sampled_params: A tensor containing samples for the uncertain
            system parameters. Note that the number of samples must be either 1
            or the number of trajectories. If 1, a single sample is used for all
            trajectories, otherwise use one sample per trajectory.
        :type sampled_params: dict
        :returns: A tensor with the next states of one or multiple trajectories.
        :rtype: th.Tensor
        """
        raise NotImplementedError

    def rejection_sampling(
        self, num_samples, x_min=-float("inf"), x_max=float("inf")
    ):
        """Samples model parameters using Rejection Sampling.

        :param num_samples: A integer with the number of output samples.
        :type num_samples: int
        :param x_min: Minimum value accepted for each sample. Same for all
            dimensions.
        :type x_min: float
        :param x_max: Maximum value accepted for each sample. Same for all
            dimensions.
        :type x_max: float
        :returns: Samples in a  2-D tensor of dimensions `n_samples` x
        `uncertain_parameters`.
        :rtype: th.Tensor
        """
        n_samples = num_samples
        dim_params = len(self.__params_keys)

        n_accepts = 0
        n_attempts = 0
        # Keeps generating numbers until we achieve the desired n
        samples = th.FloatTensor()  # output list of accepted samples
        while n_accepts < n_samples:
            # valid_samples replaces samples outside boundaries with 'nan'
            valid_samples = self.params_dist.sample([n_samples - n_accepts])
            valid_samples = valid_samples.where(
                x_min < valid_samples, th.as_tensor(float("nan"))
            )
            valid_samples = valid_samples.where(
                x_max > valid_samples, th.as_tensor(float("nan"))
            )
            # adds elements which are not nan to samples
            if dim_params == 1:
                samples = th.cat(
                    (samples, valid_samples[~th.isnan(valid_samples)]), 0
                )
                n_accepts = samples.numel()
            else:
                samples = th.cat(
                    (
                        samples,
                        valid_samples[~th.isnan(valid_samples.sum(dim=1))],
                    ),
                    0,
                )
                n_accepts = samples.numel() / dim_params
            n_attempts = n_attempts + 1

        return samples.reshape([n_samples] + [dim_params]), n_attempts

    def sample_params(
        self, num_samples, x_min=-float("inf"), x_max=float("inf")
    ):
        """Samples parameters for the forward model.

        :param num_samples: a list with the length of the parameter vector.
        Must be either [1] for a single set of parameters per time step for all
        trajectories, or [n_samples] for individual set o parameters for each
        trajectory at each time step (default: [1]).
        :param x_min: minimum accepted values for sampled parameters, same for
            all dimensions.
        :param x_max: maximum accepted values for sampled parameters, same for
            all dimensions.
        :returns: A dictionary of tensors containing the sampled parameters.
        :rtype: dict
        """
        assert (
            self.params_dist is not None
        ), "No sampling distribution specified"

        assert num_samples > 0, "Need at least one sample."

        samples, _ = self.rejection_sampling(
            num_samples, x_min=x_min, x_max=x_max
        )
        return {
            key: samples[:, idx].reshape(-1, 1)
            for (idx, key) in enumerate(self.__params_keys)
        }

    def to_params_dict(self, params):
        return {
            key: params[:, idx].reshape(-1, 1)
            for (idx, key) in enumerate(self.__params_keys)
        }



from __future__ import division

import torch as th


class MerweScaledUTF:
    """Creates a transformer object using Merwe Scaled Sigma Points [1] for the
    unscented transform. Based on the code available at [2].

    .. note:: Sigma points are parameterized using default alpha, beta, kappa
    terms, according to [2] this is currently the de-facto standard in most
    publications.

    .. [1] R. Van der Merwe "Sigma-Point Kalman Filters for Probabilistic
       Inference in Dynamic State-Space Models" (Doctoral dissertation)
    .. [2] https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python
    """

    def __init__(self, n, alpha=1e-3, beta=2, kappa=0, sqrt_method=None):
        """Constructor for the MerweScaledUTF.

        :param n: Dimensionality of the state. 2n+1 weights will be generated.
        :type n: int
        :param alpha: Defines the spread of the sigma points around the mean.
            Usually a small positive value. Defaults to 1e-3.
        :type alpha: float
        :param beta: Incorporates prior knowledge of the distribution of the
            mean. For Gaussian distributions, beta=2 is optimal. Defaults to 2.
        :type beta: float
        :param kappa: Secondary scaling parameter usually set to 0 or to 3-n.
            Defaults to 0.
        :type kappa: float
        :param sqrt_method: Defines how we compute the square root of a matrix,
            which has no unique answer. Cholesky is the default choice due to
            its speed. As of [1] this was not a well researched area and
            different choices affect how the sigma points are arranged relative
            to the eigenvectors of the covariance matrix. If the default is not
            used, the alternative method must return an upper triangular matrix
            If None, defaults to `scipy.linalg.cholesky`.
        :type sqrt_method: function
        """
        self.n = n
        self.pts = 2 * n + 1
        self.alpha = alpha
        self.beta = beta
        self.kappa = kappa
        if sqrt_method is None:
            self.sqrt = lambda x: th.cholesky(x, upper=True)
        else:
            self.sqrt = sqrt_method
        self._compute_weights()

    def __repr__(self):
        args = (
            "n: {},\n alpha: {},\n beta: {},\n kappa: {},\n"
            "loc_weights:\n{},\ncov_weights:\n{},\n".format(
                self.n,
                self.alpha,
                self.beta,
                self.kappa,
                self.__loc_weights,
                self.__cov_weights,
            )
        )
        return "{}({})".format(self.__class__.__name__, args)

    @property
    def loc_weights(self):
        """Tensor with weights for the mean computation."""
        return self.__loc_weights

    @property
    def cov_weights(self):
        """Tensor with weights for the covariance computation."""
        return self.__cov_weights

    def _compute_weights(self):
        """Computes the weights for the Merwe scaled sigma points."""
        n = self.n
        lambda_ = self.alpha ** 2 * (n + self.kappa) - n
        c = 0.5 / (n + lambda_)
        self.__cov_weights = th.ones(self.pts, dtype=th.float) * c
        self.__loc_weights = th.ones(self.pts, dtype=th.float) * c
        self.__cov_weights[0] = lambda_ / (n + lambda_) + (
            1 - self.alpha ** 2 + self.beta
        )
        self.loc_weights[0] = lambda_ / (n + lambda_)

    def compute_sigma_points(self, mu, K):
        """Computes the sigma points for a given distribution given its mean
        (mu) and covariance (K) matrix.

        :param mu: A tensor-like object of length :attr:`n` containing the
            distribution means. Can be a scalar if 1D.
        :param K: A tensor-like object of size (:attr:`n`, :attr:`n`)
            containing the covariance matrix of the distribution. If argument
            is scalar, assume form as torch.eye(n)*P.

        :return tensor sigmas : Two dimensional tensor of of size [n, 2n+1]
            containing the sigma points. Each row contains all of the sigmas
            for one dimension in the state space.
        """
        # convert all floating points to th.float64 for numerical precision
        mu = th.as_tensor(mu, dtype=th.float)
        K = th.as_tensor(K, dtype=th.float)
        if self.n != mu.size(0):
            raise ValueError(
                "expected size(x) {}, but size is {}".format(self.n, mu.size(0))
            )
        n = self.n
        lambda_ = self.alpha ** 2 * (n + self.kappa) - n
        U = self.sqrt((lambda_ + n) * K)

        sigmas = th.zeros(n, self.pts, dtype=th.float)
        sigmas[:, 0] = mu
        # Use broadcasting to compute remaining points
        sigmas[:, 1 : n + 1] = U + mu.view(-1, 1)
        sigmas[:, n + 1 : self.pts] = -U + mu.view(-1, 1)
        return sigmas

    def unscented_transform(self, sigmas):
        """Computes unscented transform of a set of sigma points using
        transformer weights. Returns the mean and covariance in a tuple.

        :param sigmas: tensor of size [n, 2n+1], with n the input dimension,
            contains the transformed sigma points
        :returns mu: tensor of size [n] with mean vector
        :returns K: tensor of size [n, n] with the covariance matrix
        """
        # new mean is just the sum of the sigmas * weight
        mu = sigmas @ self.loc_weights  # =\Sigma_(i=0)^(2n) (W^m[i]*X[i])
        # get residuals using broadcasting
        y = sigmas - mu.view(-1, 1)
        # new covariance is the sum of the outer product of the residuals
        # times the weights
        K = y @ th.diag(self.__cov_weights) @ y.t()
        return mu, K



import math

import torch as th

from disco.models.base import BaseModel
from disco.utils.spaces import Box


class CartPoleModel(BaseModel):
    """Model for the classic control problem of balancing a pole attached to
    a cart with an un-actuated joint using binary actions.

    Refer to A. G. Barto, R. S. Sutton, and C. W. Anderson, “Neuron-like
    adaptive elements that can solve difficult learning control problems,”
    IEEE Transactions on Systems, Man, and Cybernetics, vol. SMC-13,
    pp. 834–846, Sept./Oct. 1983.

    Observation space:
        Type: Box(4)

        +-----+----------------------+---------+--------+
        | Num | Observation          | Min     | Max    |
        +=====+======================+=========+========+
        | 0   | Cart Position        | -4.8    | 4.8    |
        +-----+----------------------+---------+--------+
        | 1   | Cart Velocity        | -Inf    | Inf    |
        +-----+----------------------+---------+--------+
        | 2   | Pole Angle           | -24 deg | 24 deg |
        +-----+----------------------+---------+--------+
        | 3   | Pole Velocity At Tip | -Inf    | Inf    |
        +-----+----------------------+---------+--------+

    Actions space:
        Type: Box(1)

        +-----+-----------+-----+-----+
        | Num | Action    | Min | Max |
        +=====+===========+=====+=====+
        | 0   | Push cart | -1  | +1  |
        +-----+-----------+-----+-----+
    """

    def __init__(
        self,
        g=9.8,
        f_mag=10.0,
        mass_cart=1.0,
        mass_pole=0.1,
        length=1.0,
        mu_c=0.5e-3,
        mu_p=2e-6,
        **kwargs
    ):
        """Constructor for CartPoleModel.

        :param g: Gravity in `m/s^2`. Defaults to `9.8`.
        :type g: float
        :param f_mag: Magnitude of the force applied to the cart in `N`.
            Defaults to `10.0`.
        :type f_mag: float
        :param mass_cart: Mass of the cart in `kg`. Defaults to `1.0`.
        :type mass_cart: float
        :param mass_pole: Mass of the pole in `kg`. Defaults to `0.1`.
        :type mass_pole: float
        :param length: Pole length in `m`. Defaults to `1.0`.
        :type length: float
        :param mu_c: Cart friction constant. Defaults to `0.5e(-3)`.
        :type mu_c: float
        :param mu_p: Pole friction constant. Defaults to `2e(-6)`.
        :type mu_p: float
        :key dt: Duration of each discrete update in s. (default: 0.05)
        :type dt: float
        :key uncertain_params: A tuple containing the uncertain parameters of
            the forward model. Is used as keys for assigning sampled parameters
            from the `params_dist` function. Defaults to None.
        :key params_dist: A distribution to sample parameters for the forward
            model.
        :type kwargs: tuple or torch.distributions.distribution.Distribution
        """
        super(CartPoleModel, self).__init__(**kwargs)
        self.__params_dict = {
            "g": g,
            "mass_cart": mass_cart,
            "mass_pole": mass_pole,
            "length": length,
            "mu_c": mu_c,
            "mu_p": mu_p,
            "f_mag": f_mag,
        }

        # Angle at which to fail the episode
        self.theta_threshold_radians = 12 * 2 * math.pi / 360
        self.x_threshold = 2.4

        # Angle limit set to 2 * theta_threshold_radians so failing observation
        # is still within bounds
        high = th.tensor(
            [
                self.x_threshold * 2,
                float("Inf"),
                self.theta_threshold_radians * 2,
                float("Inf"),
            ]
        )

        self.__action_space = Box(dim=1, low=-1, high=1, dtype=th.float32)
        self.__observation_space = Box(4, -high, high, dtype=th.float32)

    @property
    def observation_space(self):
        """The Pendulum observation space.

        :return: A space with the Pendulum observation space.
        :rtype: Box
        """
        return self.__observation_space

    @property
    def action_space(self):
        """The Pendulum action space.

        :return: A space with the Pendulum action space.
        :rtype: Discrete
        """
        return self.__action_space

    def step(self, states, actions, sampled_params=None):
        """Receives tensors of current states and actions and computes the
        states for the subsequent timestep. If sampled parameters are provided,
        these must be used, otherwise revert to mode of distribution over the
        uncertain parameters.

        Must be bounded by observation and action spaces.

        :param states: A tensor containing the current states of one or multiple
            trajectories.
        :type states: th.Tensor
        :param actions: A tensor containing the next planned actions of one or
            multiple trajectories.
        :type actions: th.Tensor
        :param sampled_params: A tensor containing samples for the uncertain
            system parameters. Note that the number of samples must be either 1
            or the number of trajectories. If 1, a single sample is used for all
            trajectories, otherwise use one sample per trajectory.
        :type sampled_params: dict
        :returns: A tensor with the next states of one or multiple trajectories.
        :rtype: th.Tensor
        """
        dt = self.dt
        x, x_d, theta, theta_d = states.chunk(4, dim=1)
        if sampled_params is not None:
            params_dict = self.__params_dict.copy()
            for key in sampled_params.keys():
                params_dict[key] = sampled_params[key]
            g, m_c, m_p, length, mu_c, mu_p, f_mag = params_dict.values()
        else:
            g, m_c, m_p, length, mu_c, mu_p, f_mag = self.__params_dict.values()
        # even though gym CartPole is binary, since our forward model is
        # continuous we have to center actions at 0 and clip on [-1, +1]
        actions = th.clamp(actions, min=-1, max=1)
        u = actions * f_mag

        mass = m_c + m_c  # total mass
        pm = m_p * length  # pole-mass
        cart_friction = mu_c * x_d.sign()
        pole_friction = (mu_p * theta_d) / pm
        factor = (u + pm * theta.sin() * theta_d ** 2 - cart_friction) / mass
        tdd_num = g * theta.sin() - theta.cos() * factor - pole_friction
        tdd_den = length * (4.0 / 3 - (m_p * theta.cos() ** 2) / mass)
        theta_dd = tdd_num / tdd_den

        x_dd = factor - pm * theta_dd * th.cos(theta) / mass
        delta = th.cat([x_d, x_dd, theta_d, theta_dd], dim=1) * dt
        return states + delta



# TODO: Does cost normalisation improve trajectory averaging?
# TODO: Should a decaying factor be added?
# TODO: Make controller robust to unstable rollouts
# TODO: Should the action clipping be ignored by the controller and enforced by
#   the model?

from __future__ import division

import torch as th

from disco.controllers.base import BaseController
from disco.utils.utf import MerweScaledUTF


class AMPPI(BaseController):
    """Implements an variation of the IT-MPC controller as defined in [1]_ for
    OpenAI Gym environments.

    .. [1] Williams et al., 2017 'Information Theoretic MPC for Model-Based
        Reinforcement Learning'
    """

    def __init__(
        self,
        observation_space,
        action_space,
        hz_len,
        n_samples,
        lambda_=1.0,
        a_cov=None,
        inst_cost_fn=None,
        term_cost_fn=None,
        params_sampling="extended",
        **kwargs
    ):
        """Constructor for AMPPI.

        :param observation_space: A Box object defining the action space.
        :type observation_space: gym.spaces.Space
        :param action_space: A Box object defining the action space.
        :type action_space: gym.spaces.Space
        :param hz_len: Number of time steps in the control horizon.
        :type hz_len: int
        :param lambda_: Controller regularization parameter. Defaults to 1.0.
        :param a_cov: covariance matrix of the actions multiplicative Gaussian
            noise. Effectively determines the amount of exploration
            of the system. If None, an appropriate identity matrix is used.
            Defaults to None.
        :type a_cov: th.Tensor
        :key inst_cost_fn: A function that receives a trajectory and returns
            the instantaneous cost. Must be defined if no `term_cost_fn` is
            given. Defaults to None.
        :type kwargs: function
        :key term_cost_fn: A function that receives a state
            and returns its terminal cost. Must be defined if no
            `inst_cost_fn` is given. Defaults to None.
        :param params_sampling: Can be set to either 'none, 'single',
            'extended', or a Transformer object. If 'none', mean values of the
            parameter distribution are used if available. Otherwise, default
            model parameters are used. If 'single', one sample per rollout is
            taken and used for all `n_samples` trajectories. If 'extended',
            `n_samples` are taken per rollout, meaning each trajectory has their
            own sampled parameters. Finally, if a Transformer is provided it
            will be used *instead* of sampling parameters. Defaults to
            'extended'.
        :type params_sampling: str or utils.utf.MerweScaledUTF
        :key init_actions:  A tensor of dimension `hz_len` by
            `action_space.shape` containing the initial set of control actions.
            If None, the sequence is initialized to zeros. Defaults to None.
        :type kwargs: th.Tensor

        .. note::
            * Actions will be clipped according bounding action space regardless
              of the covariance set. Effectively `epsilon <= (max_a - min_a)`.

            * Sampling is ignored when a transformer object is provided to
              method `update_actions`.
        """
        super(AMPPI, self).__init__(
            observation_space,
            action_space,
            hz_len,
            inst_cost_fn,
            term_cost_fn,
            **kwargs
        )
        self.n_samples = n_samples
        self.lambda_ = lambda_
        if a_cov is None:
            a_cov = th.eye(self.dim_a)
        a_loc = th.zeros(self.dim_a)
        self.a_dist = th.distributions.multivariate_normal.MultivariateNormal(
            a_loc, a_cov
        )
        self.a_pre = th.inverse(a_cov)

        if not params_sampling or params_sampling == "none":
            self.__params_sample_shape = None
            self.__tf = None
        elif params_sampling == "single":
            self.__params_sample_shape = 1
            self.__tf = None
        elif params_sampling == "extended":
            self.__params_sample_shape = n_samples
            self.__tf = None
        elif isinstance(params_sampling, MerweScaledUTF):
            self.__params_sample_shape = None
            self.__tf = params_sampling
        else:
            raise ValueError(
                "Invalid value for 'params_sampling': {}".format(
                    params_sampling
                )
            )
        self.__params_sampling = params_sampling

    @property
    def params_sampling(self):
        return self.__params_sampling

    def _sample_trajectories(self, model, state):
        """Sample trajectories.

        :param model: A model object which provides a `step` function to
            generate the system rollouts. If params_sampling is used, it must
            also implement a `sample_params` function for the parameters of the
            transition function.
        :type model: models.base.BaseModel
        :param state: The initial state of the system.
        :type state: th.Tensor
        :returns: A tuple of (actions, states, eps) for `n_samples` rollouts.
        :rtype: (th.Tensor, th.Tensor, th.Tensor)
        """
        # eps shape is `n_samples` x `hz_len` x `dim_a`
        eps = self.a_dist.sample(sample_shape=[self.n_samples, self.hz_len])
        actions = th.add(eps, self.a_seq)
        states = th.zeros(self.n_samples, self.hz_len + 1, self.dim_s)
        states[:, 0] = state.repeat(self.n_samples, 1)
        # shape of sampled_params is a dict with `uncertain_params` as keys and
        # tensors of size ``[(1 | `n_samples`), 1]`` samples
        if self.__params_sample_shape:
            sampled_params = model.sample_params(self.__params_sample_shape)
        else:
            sampled_params = None
        for t in range(self.hz_len):
            states[:, t + 1] = model.step(
                states[:, t], actions[:, t], sampled_params
            )
        return actions, states, eps

    def _sample_sigma_trajectories(self, model, state):
        """Sample trajectories using Unscented Transform.

        :param model: A model object which provides a `step` function to
            generate the system rollouts.
        :type model: models.base.BaseModel
        :param state: The initial state of the system.
        :type state: th.Tensor
        :returns: A tuple of (actions, states, eps) for `n_samples` * `tf.pts`
            rollouts.
        :rtype: (th.Tensor, th.Tensor, th.Tensor)
        """
        # eps shape is `n_samples` x `hz_len` x `dim_a`.
        eps = self.a_dist.sample(sample_shape=[self.n_samples, self.hz_len])
        actions = th.add(eps, self.a_seq)

        # Create a matrix to store all sigma points for each trajectory and for
        # each step, last dim is the number of sigma points
        try:
            covariance = model.params_dist.covariance_matrix
            mean = model.params_dist.mean
        except AttributeError:
            try:
                covariance = model.params_dist.variance.diag()
                mean = model.params_dist.mean
            except AttributeError:
                idx = model.params_dist.a.argmax()
                covariance = model.params_dist.xs[idx].S
                mean = model.params_dist.xs[idx].m
        params_sp = self.__tf.compute_sigma_points(mean, covariance)

        # Create tensors to be used on forward model. When using sigma points,
        # the number of trajectory is effectively `n_samples` * `tf.pts`. We'll
        # repeat `tf.pts` times each row of `actions` and `n_samples` times the
        # `params_sp` tensor, so every action is applied to each sigma point.
        acts_sp = actions.repeat(1, self.__tf.pts, 1).view(
            -1, self.hz_len, self.dim_a
        )
        sampled_params = model.to_params_dict(
            params_sp.T.repeat(self.n_samples, 1)
        )
        states_sp = th.zeros(
            self.n_samples * self.__tf.pts, self.hz_len + 1, self.dim_s
        )
        states_sp[:, 0] = state.expand_as(states_sp[:, 0])
        for t in range(self.hz_len):
            states_sp[:, t + 1] = model.step(
                states_sp[:, t], acts_sp[:, t], sampled_params
            )
        return actions, states_sp, eps

    def _compute_cost(self, states, eps):
        """Estimate trajectories cost.

        :param states: A tensor with the states of each trajectory.
        :type states: th.Tensor
        :param eps: A tensor with the difference of the current planned action
            sequence and the actions on each trajectory.
        :type eps: th.Tensor
        :returns: A tensor with the costs for the given trajectories.
        :rtype: th.Tensor
        """
        # Need to use reshape instead of view because slice is not contiguous
        inst_costs = self.inst_cost_fn(states[:, 1:].reshape(-1, self.dim_s))
        inst_costs = inst_costs.view(-1, self.hz_len).sum(dim=1)
        term_costs = self.term_cost_fn(states[:, -1])
        if self.__tf:
            # Weight trajectories using UTF sigma weights
            inst_costs = th.matmul(
                inst_costs.view(-1, self.__tf.pts), self.__tf.loc_weights
            )
            term_costs = th.matmul(
                term_costs.view(-1, self.__tf.pts), self.__tf.loc_weights
            )

        # To compute ctrl_costs in a single batch for all hz_len and all
        # n_samples, first we compute the hz_len x dim_a cost matrix, result
        # size is n_samples x hz_len x hz_len. Then take the trace of the
        # hz_len x hz_len matrices, result is dim n_samples.
        ctrl_costs = self.lambda_ * (
            th.matmul(th.matmul(self.a_seq, self.a_pre), eps.transpose(1, 2))
        ).diagonal(dim1=1, dim2=2).sum(dim=1)
        costs = term_costs + inst_costs + ctrl_costs  # shape is n_samples
        return costs

    def update_actions(self, model, state):
        """Computes the next control action and the incurred cost. Updates the
        controller next control actions.

        :param model: A model with a `step(states, actions, params)` function to
            compute the next state for a set of trajectories.
        :type model: models.base.BaseModel
        :param state: A with the system initial state.
        :type state: th.Tensor
        :returns: A tuple of `(cost, omega)` with the expected cost of the new
             actions and weights of the computed trajectories.
        :rtype: (float, th.Tensor)
        """
        state = th.as_tensor(state, dtype=th.float)
        if self.__tf:
            actions, states, eps = self._sample_sigma_trajectories(model, state)
        else:
            actions, states, eps = self._sample_trajectories(model, state)
        costs = self._compute_cost(states, eps)
        # remove nan and inf, but alters size
        finite_costs = costs.masked_select(th.isfinite(costs))
        if min(finite_costs.shape) == 0:
            print(
                "Warning: Couldn't find a feasible control. Keeping previously "
                "planned control actions."
            )
            return float("inf"), th.zeros(1)  # cost and omega
        else:
            beta = finite_costs.min()
            # beta = costs.min()
            eta = th.exp(
                (-1 / self.lambda_) * (finite_costs - beta)
            ).sum()  # scalar
            omega = (
                th.exp((-1 / self.lambda_) * (costs - beta)) / eta
            )  # tensor of size n_samples
            finite_omega = omega[th.isfinite(costs)]
            bounded_omega = omega.where(
                th.isfinite(omega), th.tensor(0.0)
            )  # replaces nan and inf by zero
            # use th.tensordot to multiply omega and epsilon for all hz_len
            self.a_seq += th.tensordot(bounded_omega, eps, dims=1)
            # Even though tau has been clipped, epsilon can be of mag
            # (u_max-u_min) so we need to clip U again
            self.a_seq = th.clamp(self.a_seq, self.min_a, self.max_a)
            cost = finite_omega.dot(finite_costs)
            return cost, states, actions, omega



import os
import pickle
import time

import gym
import torch as th

from disco.utils import distributions


def save_progress(data=None, folder_name=None, fig=None):
    """Saves session to the project data folder.

     Path can be specified, if not, an auto generated folder based on the
     current date-time is used. May include a plot.

    :param data: A data object to save. If None, whole session is saved.
    :type data: object
    :param folder_name: A path-like string containing the output path.
    :type folder_name: str
    :param fig: A figure object.
    :type fig: matplotlib.figure.Figure
    """
    if folder_name is None:
        folder_name = time.strftime("%Y%m%d-%H%M%S")
        folder_name = "data/local/" + str(folder_name)
    if not os.path.exists(os.path.join(str(folder_name), "plots")):
        os.makedirs(os.path.join(str(folder_name), "plots"))
    if fig:
        plot_path = os.path.join(str(folder_name), "plots/plot.pdf")
        fig.savefig(plot_path)

    if data is None:
        try:
            import dill
        except ImportError:
            print("Couldn't import package dill. Aborting save progress.")
            return None
        sess_path = os.path.join(str(folder_name), "session.pkl")
        dill.dump_session(sess_path)
    else:
        data_path = os.path.join(str(folder_name), "data.pt")
        th.save(data, data_path, pickle_protocol=-1)
    return folder_name


def import_mog(filename):
    """Imports a Mixture of Gaussians from a file.

    :param filename: A path-like string containing the filename to load.
    :type filename: str
    :return: A Mixture of Gaussians.
    :rtype: utils.pdf.MoG
    """
    with open(filename, "rb") as f:
        mog_dict = pickle.load(f)
        mog = distributions.MoG(
            a=mog_dict["weights"],
            ms=mog_dict["means"],
            Ss=mog_dict["covariances"],
        )
    return mog


def _plot_mean_cost(ax, acc_costs, **plot_kwargs):
    """Creates an average cost plot with standard deviation."""
    steps = acc_costs.shape[1]
    cost_mean = acc_costs.mean(dim=0)
    cost_std = acc_costs.std(dim=0) / 2
    timesteps = th.arange(steps)
    mean_ln = ax.plot(timesteps.tolist(), cost_mean.tolist(), **plot_kwargs,)
    try:
        color = plot_kwargs["color"]
        fill_ln = ax.fill_between(
            timesteps.tolist(),
            (cost_mean - cost_std).clamp(min=0.0),
            cost_mean + cost_std,
            facecolor=color,
            alpha=0.3,
        )
    except KeyError:
        fill_ln = ax.fill_between(
            timesteps.tolist(),
            (cost_mean - cost_std).clamp(min=0.0),
            cost_mean + cost_std,
            alpha=0.3,
        )
    return mean_ln, fill_ln


def run_episode(
    env_name,
    init_state,
    controller,
    model,
    binary=False,
    steps=200,
    steps_per_msg=20,
    verbose=True,
    render=False,
):
    """Runs one episode of the simulation loop.

    :param env_name: A gym environment name.
    :type env_name: str
    :param init_state: The initial state for all episodes.
    :type init_state: th.Tensor
    :param controller: The controller used for the task.
    :type controller: disco.controllers.amppi.AMPPI
    :param model: The forward model used by the controller.
    :type model: disco.models.base.BaseModel
    :param binary: Whether or not the actions should be converted to a `bool`.
    :type binary: bool
    :param steps: Number of steps in each episode.
    :type steps: int
    :param steps_per_msg: Number of steps before printing update messages.
    :type steps_per_msg: int
    :param verbose: Controls whether information messages should be print to
        screen.
    :type verbose: bool
    :param render: Enable or disable gym rendering.
    :type render: bool
    :return: A dictionary with (states, actions, cost, predicted states,
        predicted actions, omega) for each timestep.
    :rtype: dict
    """
    env = gym.make(env_name)
    env.reset()
    env.env.state = init_state  # sets the initial state of the simulation...
    state = init_state  # ... and the forward model

    # create the aggregating tensors and set them to 'nan' to check if
    # simulation breaks
    ep_dict = dict.fromkeys(
        ["states", "actions", "costs", "states_pred", "actions_pred", "omegas"]
    )

    step = 0
    while step < steps:
        if render:
            env.render()
        # values are: est. cost, states, actions, omega
        values = controller.update_actions(model, state)
        action = controller.a_seq[0]
        if binary:
            # converts to a binary scalar if CartPole
            # TODO: generalise this to other models
            action = 1 if action > 0 else 0
        controller.roll(steps=1)
        _, _, done, _ = env.step(action)
        state = th.tensor(env.state)
        cost = controller.inst_cost_fn(state.view(1, -1)).squeeze()
        if verbose and not step % steps_per_msg:
            print(
                "Step {0}: action taken {1:.2f}, cost {2:.2f}".format(
                    step, float(action), float(cost)
                )
            )
            print("Current state: {0}".format(state))
            print(
                "Next actions 4 actions: {}".format(
                    controller.a_seq[:4].flatten()
                )
            )

        #  note zip is carefully ordered
        for key, value in zip(
            ep_dict.keys(),
            [state, controller.a_seq[0], cost] + list(values[1:]),
        ):
            if step == 0:
                # initialize each dictionary entry with `nan` values of size
                # [`steps` + value.size()]
                ep_dict[key] = th.full(
                    (th.Size([steps]) + value.size()),
                    fill_value=float("nan"),
                    dtype=value.dtype,
                )
            ep_dict[key][step] = value

        if done:
            env.close()
            break
        else:
            step += 1

    if verbose:
        print(
            "Last step {0}: action taken {1:.2f}, cost {2:.2f}".format(
                step, float(action), float(cost)
            )
        )
        print("Last state: theta={0[0]}, theta_dot={0[1]}".format(state))
        print("Next actions: {}".format(controller.a_seq.flatten()))
    env.close()
    return ep_dict


def run_simulation(
    env_name,
    init_state,
    episodes,
    controller,
    model,
    binary=False,
    steps=200,
    steps_per_msg=20,
    verbose=True,
    render=False,
    plot=True,
    reuse_ax=None,
    hold=False,
    **plot_kwargs,
):
    """Runs a number of episodes in simulation and plot average cost results.

    :param env_name: A gym environment name.
    :type env_name: str
    :param init_state: The initial state for all episodes.
    :type init_state: th.Tensor
    :param episodes: Number of episodes to execute.
    :type episodes: int
    :param controller: The controller used for the task.
    :type controller: disco.controllers.amppi.AMPPI
    :param model: The forward model used by the controller.
    :type model: disco.models.base.BaseModel
    :param binary: Whether or not the actions should be converted to a `bool`.
    :type binary: bool
    :param steps: Number of steps in each episode.
    :type steps: int
    :param steps_per_msg: Number of steps before printing update messages.
    :type steps_per_msg: int
    :param verbose: Controls whether information messages should be print to
        screen.
    :type verbose: bool
    :param render: Enable or disable gym rendering.
    :type render: bool
    :param plot: Whether an average cost plot should be created.
    :type plot: bool
    :param reuse_ax: If provided, will be used to render the plot.
    :type reuse_ax: matplotlib.axes.Axes
    :param hold: Whether the plot should be displayed or held for further
        processing.
    :type hold: bool
    :param plot_kwargs: kwargs which will be passed by reference to the plot
        function.
    :type plot_kwargs: dict
    """
    import matplotlib.pyplot as plt

    sim_dict = dict()
    for episode in range(episodes):
        if verbose:
            print("=== Episode {} of {} ===".format(episode + 1, episodes))
        ep_data_dict = run_episode(
            env_name,
            init_state,
            controller,
            model,
            binary,
            steps,
            steps_per_msg,
            verbose,
            render,
        )
        if episode == 0:
            for key, value in ep_data_dict.items():
                sim_dict[key] = value.unsqueeze(0)
        else:
            for key, value in ep_data_dict.items():
                sim_dict[key] = th.cat(
                    (sim_dict[key], value.unsqueeze(0)), dim=0
                )

    if plot:
        if reuse_ax is None:
            fig, ax = plt.subplots()
        elif reuse_ax is not None:
            ax = reuse_ax
        _, _ = _plot_mean_cost(ax, sim_dict["costs"], **plot_kwargs)
        if not hold:
            ax.set_xlabel("Step")
            ax.set_ylabel(
                "Mean cost over time for {} episodes".format(episodes)
            )
            ax.legend()
            ax.grid(True, axis="y")
            ax.autoscale(enable=True, axis="x", tight=True)
            plt.show()

    return sim_dict



import numpy as np
import numpy.random as rng
import torch as th
from scipy.special import logsumexp
from scipy.stats import multivariate_normal

try:
    from itertools import izip as zip
except ImportError:  # will be 3.x series
    pass


class Gaussian:
    """Implements a gaussian pdf. Focus is on efficient multiplication, division
    and sampling.
    """

    def __init__(self, m=None, P=None, U=None, S=None, Pm=None, L=None):
        """Initialize a gaussian pdf given a valid combination of its
        parameters. Valid combinations are: m-P, m-U, m-S, Pm-P, Pm-U, Pm-S.

        :param m: mean
        :param P: precision
        :param U: upper triangular precision factor such that U'U = P
        :param S: covariance
        :param C : upper or lower triangular covariance factor, in any case
        S = C'C
        :param Pm: precision times mean such that P*m = Pm
        :param L: lower triangular covariance factor given as 1D array such that
        LL' = S
        """
        if m is not None:
            m = np.asarray(m)
            self.m = m
            self.ndim = m.size

            if P is not None:
                P = np.asarray(P)
                L = np.linalg.cholesky(P)
                self.P = P
                self.C = np.linalg.inv(L)
                self.S = np.dot(self.C.T, self.C)
                self.Pm = np.dot(P, m)
                self.logdetP = 2.0 * np.sum(np.log(np.diagonal(L)))

            elif U is not None:
                U = np.asarray(U)
                self.P = np.dot(U.T, U)
                self.C = np.linalg.inv(U.T)
                self.S = np.dot(self.C.T, self.C)
                self.Pm = np.dot(self.P, m)
                self.logdetP = 2.0 * np.sum(np.log(np.diagonal(U)))

            elif L is not None:
                L = np.asarray(L)
                Lm = np.zeros((self.ndim, self.ndim))
                idx_l = np.tril_indices(self.ndim, -1)
                idx_d = np.diag_indices(self.ndim)

                Lm[idx_l] = L[self.ndim :]
                Lm[idx_d] = L[0 : self.ndim]
                self.C = Lm.T
                self.S = np.dot(self.C.T, self.C)
                self.P = np.linalg.inv(self.S)
                self.Pm = np.dot(self.P, m)
                self.logdetP = -2.0 * np.sum(np.log(np.diagonal(self.C)))

            elif S is not None:
                S = np.asarray(S)
                self.P = np.linalg.inv(S)
                self.C = np.linalg.cholesky(S).T
                self.S = S
                self.Pm = np.dot(self.P, m)
                self.logdetP = -2.0 * np.sum(np.log(np.diagonal(self.C)))

            else:
                raise ValueError("Precision information missing.")

        elif Pm is not None:
            Pm = np.asarray(Pm)
            self.Pm = Pm
            self.ndim = Pm.size

            if P is not None:
                P = np.asarray(P)
                L = np.linalg.cholesky(P)
                self.P = P
                self.C = np.linalg.inv(L)
                self.S = np.dot(self.C.T, self.C)
                self.m = np.linalg.solve(P, Pm)
                self.logdetP = 2.0 * np.sum(np.log(np.diagonal(L)))

            elif U is not None:
                U = np.asarray(U)
                self.P = np.dot(U.T, U)
                self.C = np.linalg.inv(U.T)
                self.S = np.dot(self.C.T, self.C)
                self.m = np.linalg.solve(self.P, Pm)
                self.logdetP = 2.0 * np.sum(np.log(np.diagonal(U)))

            elif S is not None:
                S = np.asarray(S)
                self.P = np.linalg.inv(S)
                self.C = np.linalg.cholesky(S).T
                self.S = S
                self.m = np.dot(S, Pm)
                self.logdetP = -2.0 * np.sum(np.log(np.diagonal(self.C)))

            else:
                raise ValueError("Precision information missing.")

        else:
            raise ValueError("Mean information missing.")

    @property
    def mean(self):
        """Returns the Gaussian mean."""
        return self.m

    @property
    def covariance(self):
        """Returns the Gaussian covariance."""
        return self.S

    def gen(self, n_samples=1):
        """Returns independent samples from the gaussian."""
        z = rng.randn(n_samples, self.ndim)
        samples = np.dot(z, self.C) + self.m
        return samples

    def sample(self, n_samples=1):
        """Returns `n_samples` from the Gaussian in a vector."""
        samples = self.gen(n_samples)
        return samples

    def eval(self, x, ii=None, log=True):
        """Evaluates the gaussian pdf.

        :param x: rows are inputs to evaluate at
        :param ii: a list of indices specifying which marginal to evaluate. If
        None, the joint pdf is evaluated
        :param log: if True, the log pdf is evaluated
        :return: pdf or log pdf
        """
        if ii is None:
            xm = x - self.m
            lp = -np.sum(np.dot(xm, self.P) * xm, axis=1)
            lp += self.logdetP - self.ndim * np.log(2.0 * np.pi)
            lp *= 0.5

        else:
            m = self.m[ii]
            S = self.S[ii][:, ii]
            lp = multivariate_normal.logpdf(x, m, S)
            lp = np.array([lp]) if x.shape[0] == 1 else lp

        res = lp if log else np.exp(lp)
        return res

    def __mul__(self, other):
        """Multiply with another gaussian."""
        assert isinstance(other, Gaussian)
        P = self.P + other.P
        Pm = self.Pm + other.Pm
        return Gaussian(P=P, Pm=Pm)

    def __imul__(self, other):
        """Incrementally multiply with another gaussian."""
        assert isinstance(other, Gaussian)
        res = self * other
        self.m = res.m
        self.P = res.P
        self.C = res.C
        self.S = res.S
        self.Pm = res.Pm
        self.logdetP = res.logdetP
        return res

    def __truediv__(self, other):
        """Divide by another gaussian. Note that the resulting gaussian might be
        improper."""
        assert isinstance(other, Gaussian)
        P = self.P - other.P
        Pm = self.Pm - other.Pm
        return Gaussian(P=P, Pm=Pm)

    def __div__(self, other):
        """For backwards compatibility with Python 2.7"""
        return self.__truediv__(other)

    def __idiv__(self, other):
        """Incrementally divide by another gaussian. Note that the resulting
        gaussian might be improper."""
        assert isinstance(other, Gaussian)
        res = self / other
        self.m = res.m
        self.P = res.P
        self.C = res.C
        self.S = res.S
        self.Pm = res.Pm
        self.logdetP = res.logdetP
        return res

    def __pow__(self, power, modulo=None):
        """Raise gaussian to a power and get another gaussian."""
        P = power * self.P
        Pm = power * self.Pm
        return Gaussian(P=P, Pm=Pm)

    def __ipow__(self, power):
        """Incrementally raise gaussian to a power."""
        res = self ** power
        self.m = res.m
        self.P = res.P
        self.C = res.C
        self.S = res.S
        self.Pm = res.Pm
        self.logdetP = res.logdetP
        return res

    def kl(self, other):
        """Calculates the kl divergence from this to another gaussian, i.e.
        KL(this | other)."""
        assert isinstance(other, Gaussian)
        assert self.ndim == other.ndim
        t1 = np.sum(other.P * self.S)
        m = other.m - self.m
        t2 = np.dot(m, np.dot(other.P, m))
        t3 = self.logdetP - other.logdetP
        t = 0.5 * (t1 + t2 + t3 - self.ndim)
        return t


class MoG:
    """Implements a mixture of Gaussians."""

    def __init__(self, a, ms=None, Ps=None, Us=None, Ss=None, xs=None, Ls=None):
        """Creates a mog with a valid combination of parameters or an already
        given list of gaussian variables.
        :param a: mixing coefficients
        :param ms: means
        :param Ps: precisions
        :param Us: precision factors such that U'U = P
        :param Ss: covariances
        :param xs: list of gaussian variables
        :param Ls: lower-triangular covariance factor such that L*L' = S
        """
        if ms is not None:

            if Ps is not None:
                self.xs = [Gaussian(m=m, P=P) for m, P in zip(ms, Ps)]

            elif Us is not None:
                self.xs = [Gaussian(m=m, U=U) for m, U in zip(ms, Us)]

            elif Ss is not None:
                self.xs = [Gaussian(m=m, S=S) for m, S in zip(ms, Ss)]

            elif Ls is not None:
                self.xs = [Gaussian(m=m, L=L) for m, L in zip(ms, Ls)]

            else:
                raise ValueError("Precision information missing.")

        elif xs is not None:
            self.xs = xs

        else:
            raise ValueError("Mean information missing.")

        self.a = np.asarray(a)
        self.ndim = self.xs[0].ndim
        self.n_components = len(self.xs)
        self.ncomp = self.n_components
        self.__gaussian = self.project_to_gaussian()

    @property
    def mean(self):
        """Returns the mean of the projected Gaussian."""
        return self.__gaussian.m

    @property
    def covariance(self):
        """Returns the covariance of the projected Gaussian"""
        return self.__gaussian.S

    def gen(self, n_samples=1):
        """Generates independent samples from mog."""
        ii = discrete_sample(self.a, n_samples)
        ns = [np.sum((ii == i).astype(int)) for i in range(self.n_components)]
        samples = [x.gen(n_samples=n) for x, n in zip(self.xs, ns)]
        samples = np.concatenate(samples, axis=0)
        # Uncomment the lines below another possible way of sampling from MoG
        # weighted_sample = [sample * weight for sample, weight in zip(samples,
        # self.a)]
        # sample = sum(weighted_sample)
        return samples

    def sample(self, sample_shape=(1,)):
        """Returns `n_samples` from the MoG in a tensor."""
        samples = self.gen(sample_shape[0])
        return th.as_tensor(samples, dtype=th.float)

    def eval(self, x, ii=None, log=True):
        """Evaluates the mog pdf.

        :param x: rows are inputs to evaluate at
        :param ii: a list of indices specifying which marginal to evaluate. if
        None, the joint pdf is evaluated
        :param log: if True, the log pdf is evaluated
        :return: pdf or log pdf
        """
        ps = np.array(
            [self.xs[ix].eval(x, ii, log) for ix in range(len(self.a))]
        ).T
        res = (
            logsumexp(ps + np.log(self.a), axis=1)
            if log
            else np.dot(ps, self.a)
        )
        return res

    def __mul__(self, other):
        """Multiply with a single gaussian."""
        assert isinstance(other, Gaussian)
        ys = [x * other for x in self.xs]
        lcs = np.empty_like(self.a)
        for i, (x, y) in enumerate(zip(self.xs, ys)):
            lcs[i] = x.logdetP + other.logdetP - y.logdetP
            lcs[i] -= (
                np.dot(x.m, np.dot(x.P, x.m))
                + np.dot(other.m, np.dot(other.P, other.m))
                - np.dot(y.m, np.dot(y.P, y.m))
            )
            lcs[i] *= 0.5
        la = np.log(self.a) + lcs
        la -= logsumexp(la)
        a = np.exp(la)
        return MoG(a=a, xs=ys)

    def __imul__(self, other):
        """Incrementally multiply with a single gaussian."""
        assert isinstance(other, Gaussian)
        res = self * other
        self.a = res.a
        self.xs = res.xs
        return res

    def __truediv__(self, other):
        """Divide by a single gaussian."""
        assert isinstance(other, Gaussian)
        ys = [x / other for x in self.xs]
        lcs = np.empty_like(self.a)
        for i, (x, y) in enumerate(zip(self.xs, ys)):
            lcs[i] = x.logdetP - other.logdetP - y.logdetP
            lcs[i] -= (
                np.dot(x.m, np.dot(x.P, x.m))
                - np.dot(other.m, np.dot(other.P, other.m))
                - np.dot(y.m, np.dot(y.P, y.m))
            )
            lcs[i] *= 0.5
        la = np.log(self.a) + lcs
        la -= logsumexp(la)
        a = np.exp(la)
        return MoG(a=a, xs=ys)

    def __div__(self, other):
        """For backwards compatibility with Python 2.7."""
        return self.__truediv__(other)

    def __idiv__(self, other):
        """Incrementally divide by a single gaussian."""
        assert isinstance(other, Gaussian)
        res = self / other
        self.a = res.a
        self.xs = res.xs
        return res

    def _get_summary_stats_from_mog(self):
        """Computes the mean and covariance of a MoG.

        :returns (mu, S): x is a tensor of size [n] with the grand mean and P a
        tensor of size [n, n] with the covariances.
        """
        alpha = self.a
        res_mean = np.zeros((self.ncomp, self.ndim))
        mu = np.zeros(self.ndim)
        S = np.zeros((self.ndim, self.ndim))
        # We'll decompose the MoG into x = y + N(0, S), where y will be the
        # weighted mean and the covariance is given by the law of total variance
        # S = E(var(x∣y))+var(E(x∣y))
        for idx, gauss in enumerate(self.xs):
            mu += alpha[idx] * gauss.m
            S += alpha[idx] * gauss.C  # first term of covariance matrix
            res_mean[idx] = gauss.m
        res_mean -= mu
        S += np.matmul(
            (alpha * res_mean.T), res_mean
        )  # add second term of covariance
        return mu, S

    def project_to_gaussian(self):
        """Returns a gaussian with the same mean and precision as self."""
        m, S = self._get_summary_stats_from_mog()
        return Gaussian(m, S)

    def prune_negligible_components(self, threshold):
        """Removes all the components whose mixing coefficient is less than a
        threshold."""
        ii = np.nonzero((self.a < threshold).astype(int))[0]
        total_del_a = np.sum(self.a[ii])
        del_count = ii.size
        self.n_components -= del_count
        self.a = np.delete(self.a, ii)
        self.a += total_del_a / self.n_components
        self.xs = [x for i, x in enumerate(self.xs) if i not in ii]

    def kl(self, other, n_samples=10000):
        """Estimates the kl from this to another pdf, i.e. KL(this | other),
        using monte carlo."""
        x = self.gen(n_samples)
        lp = self.eval(x, log=True)
        lq = other.eval(x, log=True)
        t = lp - lq
        res = np.mean(t)
        err = np.std(t, ddof=1) / np.sqrt(n_samples)
        return res, err


def discrete_sample(p, n_samples=1):
    """Samples from a discrete distribution.

    :param p: a distribution with N elements
    :param n_samples: number of samples
    :return: vector of samples """
    # cumulative distribution
    c = np.cumsum(p[:-1])[np.newaxis, :]
    # get the samples
    r = rng.rand(n_samples, 1)
    return np.sum((r > c).astype(int), axis=1)



import math

import torch as th

from disco.models.base import BaseModel
from disco.utils.spaces import Box


class PendulumModel(BaseModel):
    """Model for a fixed joint 1-DOF inverted pendulum.

    For more information refer to OpenAI Gym environment at:
    https://gym.openai.com/envs/Pendulum-v0/
    """

    def __init__(self, g=9.8, mass=1.0, length=1.0, **kwargs):
        """Constructor for PendulumModel.

        :param g: Gravity in m/s^2. Defaults to 9.8.
        :param mass: Pendulum mass in kg. Defaults to 1.0.
        :param length: Pendulum length in metres. Defaults to 1.0.
        :key dt: Duration of each discrete update in s. Defaults to 0.05
        :type dt: float
        :key uncertain_params: A tuple containing the uncertain parameters of
            the forward model. Is used as keys for assigning sampled parameters
            from the `params_dist` function. Defaults to None.
        :key params_dist: A distribution to sample parameters for the forward
            model.
        :type kwargs: tuple or torch.distributions.distribution.Distribution
        """
        super(PendulumModel, self).__init__(**kwargs)
        self.__max_speed = 8.0
        self.__max_torque = 2.0
        bounds = th.tensor([float("inf"), self.__max_speed])
        self.__observation_space = Box(
            dim=2, low=-bounds, high=bounds, dtype=th.float32
        )
        self.__action_space = Box(
            dim=1,
            low=-self.__max_torque,
            high=self.__max_torque,
            dtype=th.float32,
        )
        # `__params_dict` is a dictionary with default parameters values
        self.__params_dict = {"g": g, "mass": mass, "length": length}
        if self.uncertain_params is not None:
            for (idx, key) in enumerate(self.uncertain_params):
                self.__params_dict[key] = self.params_dist.mean[idx]

    @property
    def observation_space(self):
        """The Pendulum observation space.

        :return: A space with the Pendulum observation space.
        :rtype: Box
        """
        return self.__observation_space

    @property
    def action_space(self):
        """The Pendulum action space.

        :return: A space with the Pendulum action space.
        :rtype: Box
        """
        return self.__action_space

    def step(self, states, actions, sampled_params=None):
        """Receives tensors of current states and actions and computes the
        states for the subsequent timestep. If sampled parameters are provided,
        these must be used, otherwise revert to mode of distribution over the
        uncertain parameters.

        Must be bounded by observation and action spaces.

        :param states: A tensor containing the current states of one or multiple
            trajectories.
        :type states: th.Tensor
        :param actions: A tensor containing the next planned actions of one or
            multiple trajectories.
        :type actions: th.Tensor
        :param sampled_params: A tensor containing samples for the uncertain
            system parameters. Note that the number of samples must be either 1
            or the number of trajectories. If 1, a single sample is used for all
            trajectories, otherwise use one sample per trajectory.
        :type sampled_params: dict
        :returns: A tensor with the next states of one or multiple trajectories.
        :rtype: th.Tensor
        """
        dt = self.dt
        # Assigning states and params, keeping their dims
        theta, theta_d = states.chunk(2, dim=-1)
        if sampled_params is not None:
            params_dict = self.__params_dict.copy()
            for key in sampled_params.keys():
                params_dict[key] = sampled_params[key]
            g, m, length = params_dict.values()
        else:
            g, m, length = self.__params_dict.values()

        actions = th.clamp(
            actions, min=-self.__max_torque, max=self.__max_torque
        )
        theta_d = theta_d + dt * (
            -3 * g / (2 * length) * (theta + math.pi).sin()
            + 3.0 / (m * length ** 2) * actions
        )
        theta = theta + theta_d * dt  # Use new theta_d
        theta_d = th.clamp(theta_d, -self.__max_speed, self.__max_speed)

        return th.cat((theta, theta_d), dim=-1)

    @staticmethod
    def get_obs(state):
        try:
            theta, theta_d = state.chunk(2, dim=1)
        except ValueError:
            raise ValueError("Dimension 1 of state tensor must be exactly 2.")
        return th.cat([th.cos(theta), th.sin(theta), theta_d], dim=1)



import setuptools

with open("LICENSE", "r") as fh:
    license = fh.read()

with open("README.md", "r") as fh:
    readme = fh.read()

short_description = (
    "This package provides a collection of modules and examples to quickly "
    "setup experiments using DISCO: Double Likelihood-free Inference "
    "Stochastic Control."
)
setuptools.setup(
    name="disco-rl",
    version="0.0.1",
    author="Lucas Barcelos",
    author_email="lucas.barcelos@sydney.edu.au",
    description=short_description,
    long_description=readme,
    long_description_content_type='text/markdown',
    url="https://github.com/lubaroli/disco",
    packages=setuptools.find_packages(exclude=("notebooks", "docs", "scripts")),
    license="GPL",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "gym >= 0.12.4",
        "numpy >= 1.18.1",
        "matplotlib >= 3.1.2",
        "scipy >= 1.4.1",
        "torch >= 1.3",
    ]
)



import torch as th


class BaseController:
    """Base class for module controllers."""

    def __init__(
        self,
        observation_space,
        action_space,
        hz_len,
        inst_cost_fn=None,
        term_cost_fn=None,
        init_actions=None,
    ):
        """Constructor for BaseController.

        :param observation_space: A Box object defining the action space.
        :type observation_space: disco.utils.spaces.Space
        :param action_space: A Box object defining the action space.
        :type action_space: disco.utils.spaces.Space
        :param hz_len: Number of time steps in the control horizon.
        :type hz_len: int
        :param init_actions:  A tensor of dimension `hz_len` by `dim_a`
            containing the initial set of control actions. If None, the sequence
            is initialized to zeros. Defaults to None.
        :type init_actions: th.Tensor
        """
        self.hz_len = hz_len
        self.dim_s = observation_space.dim
        self.dim_a = action_space.dim
        self.min_a = float(action_space.low)
        self.max_a = float(action_space.high)
        if not init_actions:
            self.a_seq = th.zeros((self.hz_len, self.dim_a))
        else:
            self.a_seq = init_actions

        # Need at least one cost function
        if inst_cost_fn is None and term_cost_fn is None:
            raise ValueError("Specify at least one cost function")

        def _null_cost_fn(state):
            """Pseudo-function to return a null cost.

            :param state: The current state of the system.
            :type state: th.Tensor
            """
            return (state * 0).sum(-1)

        if inst_cost_fn is None:
            self.__inst_cost_fn = _null_cost_fn
        else:
            self.__inst_cost_fn = inst_cost_fn
        if term_cost_fn is None:
            self.__term_cost_fn = _null_cost_fn
        else:
            self.__term_cost_fn = term_cost_fn

    @property
    def inst_cost_fn(self):
        return self.__inst_cost_fn

    @property
    def term_cost_fn(self):
        return self.__term_cost_fn

    def roll(self, steps=1):
        """Shifts the control sequence forward in-place a given amount of
        `steps`.
        
        Shifts the current control sequence forward in-place by an amount of
        steps specified by the :code:`steps` parameter. The new rows are
        initialized to zeros.

        :param steps: The number of time steps to roll forward. Defaults to 1.

        """
        self.a_seq = th.roll(self.a_seq, -steps, 0)
        self.a_seq[-steps:, :] = th.zeros_like(self.a_seq[-steps:, :])

    @staticmethod
    def get_jacobian(inputs, outputs=None, func=None, create_graph=False):
        """Computes the Jacobian of outputs with respect to inputs.
        
        The Jacobian is computed either with respect to an output tensor or by
        computing it using the given mapping `func`. *Exactly* one of
        `outputs` or `func` arguments must be provided.

        :param inputs: A tensor for the input of param function, must have
        gradient information if transition function is not provided.
        :param outputs: A tensor with the output of `func`, must have gradient
                information. Defaults to None.
        :param func: A transition function to generate the output tensor.
            Defaults to None.
        :param create_graph: If True keeps the computational graph of the
            resulting Jacobian. Defaults to False.
        :returns: A tensor of shape (outputs.shape + inputs.shape) containing
            the Jacobian of outputs with respect to inputs.

        """
        if func is None and outputs is None:
            raise ValueError("Provide an output value or specify a function")
        if func is not None:
            inputs = inputs.squeeze().requires_grad_(True)
            outputs = func(inputs)
        jac = th.zeros(outputs.view(-1).shape + inputs.view(-1).shape)
        for i, out in enumerate(outputs.view(-1)):
            row_i = th.autograd.grad(
                out,
                inputs,
                create_graph=create_graph,
                retain_graph=True,
                allow_unused=True,
            )[0]
            if row_i is None:
                # this element of output doesn't depend on the inputs, so leave
                # gradient 0
                continue
            else:
                jac[i] = row_i
        if create_graph:
            jac.requires_grad_()
        return jac

    @staticmethod
    def get_hessian(inputs, jacobian):
        """Computes the hessian of outputs with respect to inputs.

        :param inputs: The input tensor.
        :param jacobian: A jacobian matrix with a computational graph.
        :returns: A tensor of same shape as `jacobian` containing the Hessian of
                outputs with respect to inputs.

        """
        assert jacobian.requires_grad and inputs.requires_grad
        hes = th.zeros_like(jacobian)
        for i, out in enumerate(jacobian):
            for j in range(inputs.shape[0]):
                e = th.autograd.grad(
                    out[j], inputs, retain_graph=True, allow_unused=True
                )[0]
                if e is None:
                    # this element of output doesn't depend on the inputs, so
                    # leave gradient 0
                    continue
                else:
                    hes[i, j] = e[j]
        return hes

    def update_actions(self, model, state):
        raise NotImplementedError("Should be implemented by the subclass")
